import os
import sys
import platform
import stat
import re
import shutil
from pathlib import Path
from typing import Union, Optional

#############################################################################################################

def normPath(
    string: Union[str, Path],
    pathType: Optional[str] = None,
    trailingSlash: Optional[bool] = None
):
    """
    Normalize path string
    """
    try:
        if str(string).strip() == '':
            raise
        path = Path(string)#.resolve()

    except:
        return None

    else: #if re.search(r':[/\\\\]', str(string)) or re.search(r'\./', str(string)):
        if trailingSlash is None:
            trailingSlash = True if str(string).endswith(('/', '\\')) else False
        if platform.system() == 'Windows' or pathType == 'Win32':
            string = path.as_posix().replace(r'/', '\\')
            string += '\\' if trailingSlash else ''
        if platform.system() == 'Linux' or pathType == 'Posix':
            string = path.as_posix()
            string += '/' if trailingSlash else ''
        return string

#############################################################################################################

"""
def getPaths(
    directory,
    extensions
):
    matchedFiles = []

    for extension in extensions:
        matchedFiles.extend(Path(directory).glob(f'*{extension}'))

    return matchedFiles
"""


def getPaths(
    directory: str,
    name: str,
    searchKeyword: bool = True
):
    """
    Get all paths of files and folders in directory
    """
    results = []

    for dirPath, folderNames, fileNames in os.walk(directory):
        for folderName in folderNames:
            if name == folderName or (name in folderName and searchKeyword is True):
                results.append(Path(dirPath).joinpath(folderName))
            else:
                pass
        for fileName in fileNames:
            if name == fileName or (name in fileName and searchKeyword is True):
                results.append(Path(dirPath).joinpath(fileName))
            else:
                pass

    return [result.as_posix() for result in results] if len(results) > 0 else None

#############################################################################################################

def getBaseDir(
    filePath: Optional[str] = None,
    parentLevel: Optional[int] = None,
    searchMEIPASS: bool = False
):
    """
    Get the parent directory of file, or get the MEIPASS if file is compiled with pyinstaller
    """
    if filePath is not None:
        baseDir = normPath(Path(str(filePath)).absolute().parents[parentLevel if parentLevel is not None else 0])
    elif searchMEIPASS and getattr(sys, 'frozen', None):
        baseDir = normPath(sys._MEIPASS)
    else:
        baseDir = None

    return baseDir


def getCurrentPath():
    """
    Get the path of the current python file
    """
    return Path(sys.argv[0]).as_posix() #return __file__ if getFileInfo()[1] == False else sys.executable


def getFileInfo(
    file: Optional[str] = None
):
    """
    Check whether python file is compiled
    """
    if file is None:
        fileName = Path(getCurrentPath()).name
        if getattr(sys, 'frozen', None):
            isFileCompiled = True
        else:
            isFileCompiled = False if fileName.endswith('.py') or sys.executable.endswith('python.exe') else True
    else:
        fileName = Path(normPath(file)).name
        isFileCompiled = False if fileName.endswith('.py') else True

    return fileName, isFileCompiled

#############################################################################################################

def renameIfExists(
    path: str
):
    """
    If path already exists, rename it to path(0), path(1), etc.
    """
    parentPath, name = Path(path).parent, Path(path).name
    suffix = Path(name).suffix
    if len(suffix) > 0:
        while Path(path).exists():
            pattern = r'(\d+)\)\.'
            if re.search(pattern, name) is None:
                name = name.replace('.', '(0).')
            else:
                CurrentNumber = int(re.findall(pattern, name)[-1])
                name = name.replace(f'({CurrentNumber}).', f'({CurrentNumber + 1}).')
            path = parentPath.joinpath(name).as_posix()
    else:
        while Path(path).exists():
            pattern = r'(\d+)\)'
            match = re.search(pattern, name)
            if match is None:
                name += '(0)'
            else:
                CurrentNumber = int(match.group(1))
                name = name[:match.start(1)] + f'({CurrentNumber + 1})'
            path = parentPath.joinpath(name).as_posix()
    return path


def rmtree(
    path: str
):
    """
    Remove file or directory tree including read-only files
    """
    shutil.rmtree(
        path,
        onerror = lambda func, path, execinfo: (
            os.chmod(path, stat.S_IWUSR),
            func(path)
        )
    )


def cleanDirectory(
    directory: str,
    whiteList: list
):
    """
    Remove all files and folders in directory except those in whiteList
    """
    if Path(directory).exists():
        for dirPath, folders, files in os.walk(directory, topdown = False):
            for file in files:
                filePath = Path(dirPath).joinpath(file).as_posix()
                try:
                    if not any(file in filePath for file in whiteList):
                        os.remove(filePath)
                except:
                    pass
            for folder in folders:
                folderPath = Path(dirPath).joinpath(folder).as_posix()
                try:
                    if not any(folder in folderPath for folder in whiteList):
                        rmtree(folderPath)
                except:
                    pass


def moveFiles(
    directory: str,
    destination: str
):
    """
    Move all files and folders in directory to destination
    """
    for dirPath, folderNames, fileNames in os.walk(directory):
        for folderName in folderNames:
            if directory != destination:
                shutil.move(Path(dirPath).joinpath(folderName).as_posix(), destination)
        for fileName in fileNames:
            if directory != destination:
                shutil.move(Path(dirPath).joinpath(fileName).as_posix(), destination)

#############################################################################################################


from typing import Optional, Type, Callable, Any
from functools import singledispatch, wraps

#############################################################################################################

class singledispatchmethod:
    '''
    Single-dispatch generic method descriptor.

    Supports wrapping existing descriptors and handles non-descriptor callables as instance methods.
    '''
    def __init__(self, method: Callable):
        if not (callable(method) and hasattr(method, "__get__")):
            raise TypeError(f"{method!r} is not callable or a descriptor")

        self.method = method

        self.dispatcher = singledispatch(method)

    def register(self, cls: Type, method: Optional[Callable] = None):
        '''
        Register a new implementation for the given class on a generic method.

        :param cls: The class to register the method for.
        :param method: The method to register. If None, returns a decorator.
        :return: A decorator function if method is None, otherwise None.
        '''
        return self.dispatcher.register(cls, func = method)

    @property
    def isAbstractMethod(self) -> bool:
        '''
        Check if the method is an abstract method.
        '''
        return getattr(self.method, '__isabstractmethod__', False)

    def __get__(self, obj: Any, cls: Optional[Type] = None):
        @wraps(self.method)
        def method(*args, **kwargs):
            '''
            Ref: https://stackoverflow.com/questions/24601722
            '''
            if args:
                method = self.dispatcher.dispatch(args[0].__class__)
            else:
                method = self.method
                for v in kwargs.values():
                    if v.__class__ in self.dispatcher.registry:
                        method = self.dispatcher.dispatch(v.__class__)
                        if method is not self.method:
                            break
            return method.__get__(obj, cls)(*args, **kwargs)
        method.register = self.register
        method.__isabstractmethod__ = self.isAbstractMethod
        return method

#############################################################################################################


import os
import locale
import codecs
import platform
import re
import string
import random
import unicodedata
import json
import polars
import urllib.parse as urlparse
from typing import Optional

#############################################################################################################

def getSystemEncoding(
    default = 'utf-8'
):
    """
    Get current system's encoding
    """
    try:
        # Get system encoding
        encoding = locale.getpreferredencoding(do_setlocale = False).lower()
        # Handle empty or invalid encodings
        if encoding in ['ansi_x3.4-1968', 'us-ascii', 'c', 'posix']:
            for env_var in ['LC_ALL', 'LC_CTYPE', 'LANG']:
                lang_env = os.environ.get(env_var)
                if lang_env and '.' in lang_env:
                    encoding_env = lang_env.split('.')[-1]
                    if encoding_env and encoding_env.upper() != 'UTF-8':
                        if '@' in encoding_env:
                            encoding_env = encoding_env.split('@')[0]
                        encoding = encoding_env.lower()
                        break
        encodingMap  = {
            'utf8': 'utf-8',
            'utf16': 'utf-16',
            'utf32': 'utf-32',
            'latin1': 'latin-1',
            'cp936': 'gbk',
            'cp950': 'big5',
            'cp65001': 'utf-8',
            'cp1252': 'windows-1252',
            'iso8859_1': 'latin-1',
        }
        encoding = encodingMap.get(encoding, encoding)
        # Verify if the encoding is available
        codecs.lookup(encoding)

    except (LookupError, TypeError, AttributeError, ValueError):
        # Use default encoding
        encoding = default

    return encoding

#############################################################################################################

def evalString(string: str):
    try:
        return eval(string)
    except:
        return string

#############################################################################################################

def removeLF(
    string: str,
    removeAll: bool = False,
):
    lfChar = r'\r\n' if platform.system() == 'Windows' else r'\n'
    return string.strip(lfChar) if removeAll else re.sub(lfChar, '', string)


def rawString(
    text: str
):
    """
    Return as raw string representation of text
    """
    RawMap = {
        7: r'\a',
        8: r'\b',
        9: r'\t',
        10: r'\n',
        11: r'\v',
        12: r'\f',
        13: r'\r'
    }
    text = r''.join([RawMap.get(ord(Char), Char) for Char in text])
    '''
    StringRepresentation = repr(text)[1:-1] #StringRepresentation = text.encode('unicode_escape').decode()
    return re.sub(r'\\+', lambda arg: r'\\', StringRepresentation).replace(r'\\', '\\').replace(r'\'', '\'') #return eval("'%s'" % canonical_string)
    '''
    return unicodedata.normalize('NFKC', text)

#############################################################################################################

def findURL(
    string: str
):
    """
    Function to find URL in a string
    """
    URLList = re.compile(r'http[s]?://(?:[a-zA-Z]|[0-9]|[$-_@.&+]|(?:%[0-9a-fA-F][0-9a-fA-F])|[!*\\(\\),])+').findall(rawString(string))
    return URLList[0] if URLList.__len__() < 2 else URLList


def makeSafeForURL(string):
    return urlparse.quote(
        string = str(string),
        safe = ':,[]'
    )


def isURL(content: str):
    """
    Check if content is a url
    """
    if urlparse.urlparse(content).scheme in ['http', 'https']:
        return True
    else:
        return False

#############################################################################################################

def isJson(content: str):
    """
    Check if content is a json
    """
    try:
        json.loads(json.dumps(eval(content)))
        return True
    except:
        return False

#############################################################################################################

def generateRandomString(
    amount: int = 9
):
    randomString = ''.join(
        random.sample(
            string.ascii_letters + string.digits,
            k = amount
        )
    )
    return randomString

#############################################################################################################

@polars.Config( 
    tbl_formatting = "ASCII_MARKDOWN",        
    tbl_hide_column_data_types = True,
    tbl_hide_dataframe_shape = True,
)
def _toMarkdown(df: polars.DataFrame) -> str:
    return str(df)


def toMarkdown(content: str):
    """
    Convert content to markdown
    """
    if isURL(content):
        content = f"[URL]({content})"
    if isJson(content):
        content = _toMarkdown(polars.DataFrame(json.loads(json.dumps(eval(content)))))
    return content

#############################################################################################################

class richTextManager:
    """
    Manage rich text
    """
    def __init__(self):
        self.richTextLines = []

    def _toHtml(self, text, align, size, weight, letterSpacing, lineHeight):
        Style = f"'text-align:{align}; font-size:{size}pt; font-weight:{weight}; letter-spacing: {letterSpacing}px; line-height: {lineHeight}px'"
        content = re.sub(
            pattern = "[\n]",
            repl = "<br>",
            string = text
        ) if text is not None else None
        return f"<p style={Style}>{content}</p>" if content is not None else ''

    def addTitle(self,
        text: Optional[str] = None,
        align: str = "left",
        size: float = 12.3,
        weight: float = 630.,
        spacing: float = 0.9,
        lineHeight: float = 24.6,
    ):
        head = f"<body>{self._toHtml(text, align, size, weight, spacing, lineHeight)}</body>" #head = f"<head><title>{self._toHtml(text, align, size, weight, spacing, lineHeight)}</title></head>"
        self.richTextLines.append(head)
        return self

    def addBody(self,
        text: Optional[str] = None,
        align: str = "left",
        size: float = 9.3,
        weight: float = 420.,
        spacing: float = 0.6,
        lineHeight: float = 22.2,
    ):
        body = f"<body>{self._toHtml(text, align, size, weight, spacing, lineHeight)}</body>"
        self.richTextLines.append(body)
        return self

    def richText(self):
        richText = "<html>\n%s\n</html>" % '\n'.join(self.richTextLines)
        return (richText)


def setRichText(
    text: str = "",
    align: str = "left",
    size: float = 9.6,
    weight: float = 450.,
    spacing: float = 0.66,
    lineHeight: float = 22.2,
):
    """
    Function to set rich text
    """
    return richTextManager().addBody(text, align, size, weight, spacing, lineHeight).richText()

#############################################################################################################


import os
import psutil
import signal
from concurrent.futures import ProcessPoolExecutor, ThreadPoolExecutor, Future
from enum import Enum
from typing import Union, Optional

from .path import normPath

#############################################################################################################

class taskAccelerationManager(Enum):
    """
    Manage task acceleration
    """
    ThreadPool = 0
    ProcessPool = 1

    def create(self, 
        funcDict: dict,
        asynchronous: bool = True,
        maxWorkers: Optional[int] = None
    ) -> list[Future]:
        if self == self.ThreadPool:
            executor = ThreadPoolExecutor(maxWorkers)
        if self == self.ProcessPool:
            executor = ProcessPoolExecutor(maxWorkers)
        futures = []
        for func, args in funcDict.items():
            try:
                future = executor.submit(func, *args)
                isLastFuture = len(futures) == len(funcDict) - 1
                futures.append(future) if asynchronous or isLastFuture else future.result()
            except Exception as e:
                executor.shutdown(wait = False, cancel_futures = True)
                raise e
        return futures

#############################################################################################################

def terminateProcess(
    program: Union[str, int],
    selfIgnored: bool = True,
    searchKeyword: bool = False
):
    """
    Kill a process by its PID or name
    """
    if isinstance(program, int):
        PID = program
        try:
            Process = psutil.Process(PID)
        except psutil.NoSuchProcess: # Process already terminated
            return

        ProcessList =  Process.children(recursive = True) + [Process]
        for Process in ProcessList:
            try:
                if Process.pid == os.getpid() and selfIgnored:
                    continue
                os.kill(Process.pid, signal.SIGTERM)
            except:
                pass

    if isinstance(program, str):
        name = program
        programPath = normPath(name) if normPath(name) is not None else name
        for Process in psutil.process_iter():
            ProcessList =  Process.children(recursive = True) + [Process]
            try:
                for Process in ProcessList:
                    if Process.pid == os.getpid() and selfIgnored:
                        continue
                    ProcessPath = Process.exe()
                    if programPath == ProcessPath or (programPath.lower() in ProcessPath.lower() and searchKeyword):
                        Process.send_signal(signal.SIGTERM) #Process.kill()
            except:
                pass


def terminateOccupation(
    file: str,
    searchKeyword: bool = False
):
    """
    Terminate all processes that are currently using the file
    """
    filePath = normPath(file) if normPath(file) is not None else file
    for Process in psutil.process_iter():
        try:
            PopenFiles = Process.open_files()
            for PopenFile in PopenFiles:
                PopenFilePath = PopenFile.path
                if filePath == PopenFilePath or (filePath.lower() in PopenFilePath.lower() and searchKeyword):
                    Process.send_signal(signal.SIGTERM) #Process.kill()
        except:
            pass

#############################################################################################################


from .utils import toIterable, itemReplacer, findKey, getNamesFromMethod, getClassFromMethod, runEvents
from .overload import singledispatchmethod
from .math import getDecimalPlaces
from .text import getSystemEncoding, evalString, removeLF, rawString, findURL, makeSafeForURL, isURL, isJson, generateRandomString, toMarkdown, richTextManager, setRichText
from .path import normPath, getPaths, getBaseDir, getCurrentPath, getFileInfo, renameIfExists, rmtree, cleanDirectory, moveFiles
from .log import loggerLevel, loggerManager
from .process import taskAccelerationManager, terminateProcess, terminateOccupation
from .cmd import subprocessManager, runCMD, asyncSubprocessManager, mkPyFileCommand, runScript, bootWithScript
from .env import isVersionSatisfied, isSystemSatisfied, envType, setEnvVar
from .config import configManager
from .database import sqliteManager
from .web import isPortAvailable, findAvailablePorts, freePort, requestManager, simpleRequest, responseParser, downloadFromURL, checkUpdateFromGithub


import hashlib
import polars
import sqlalchemy
from datetime import datetime

#############################################################################################################

class sqliteManager:
    """
    Manage sqlite database
    """
    historydbName = "history"
    historydbTableName = "historyTable"

    filedbName = None

    colName_fileHash = "fileHash"
    colName_filedbName = "filedbName"

    def writeDataToFiledb(self, df: polars.DataFrame, new: bool = True):
        '''
        将DataFrame导入文件数据库
        '''
        self.filedbName = f"DB_{datetime.now().strftime('%Y%m%d%H%M%S')}" if new else self.filedbName
        filedb_engine = sqlalchemy.create_engine(f'sqlite:///{self.filedbName}.db')
        df.write_database(
            table_name = self.filedbName,
            connection = filedb_engine,
            if_table_exists = 'replace'
        )

    def readDataFromFiledb(self):
        '''
        从文件数据库中导出DataFrame
        '''
        filedb_engine = sqlalchemy.create_engine(f'sqlite:///{self.filedbName}.db') # 与文件数据库建立连接
        df = polars.read_database(
            f"SELECT * FROM {self.filedbName}",
            connection = filedb_engine
        )
        df.fill_nan("")
        return df

    def createHistorydb(self):
        '''
        创建历史记录数据库并初始化历史记录Table
        '''
        self.historyEngine = sqlalchemy.create_engine(f'sqlite:///{self.historydbName}.db')
        if not sqlalchemy.inspect(self.historyEngine).has_table(self.historydbTableName):
            df = polars.DataFrame({
                self.colName_fileHash: [],
                self.colName_filedbName: []
            })
            df.write_database(
                table_name = self.historydbTableName,
                connection = self.historyEngine,
                if_table_exists = 'replace'
            )

    def toHistorydb(self, file_path):
        '''
        将[表格哈希值,表格数据库名]写入历史记录数据库
        '''
        fileHash = hashlib.md5(open(file_path, 'rb').read()).hexdigest()
        df = polars.DataFrame({
            self.colName_fileHash: [fileHash],
            self.colName_filedbName: [self.filedbName]
        })
        df.write_database(
            table_name = self.historydbTableName,
            connection = self.historyEngine,
            if_table_exists = 'append'
        )
        # TODO 通过redis建立旁路缓存模式

    def chkHistorydb(self, file_path):
        '''
        检查文件哈希值在历史记录数据库中的对应值
        '''
        fileHash = hashlib.md5(open(file_path, 'rb').read()).hexdigest()
        df = polars.read_database(
            f"SELECT * FROM {self.historydbTableName} WHERE {self.colName_fileHash} = '{fileHash}'",
            connection = self.historyEngine
        )
        filedbName = df.row(0, named = True)[self.colName_filedbName] if len(df) > 0 else None
        return filedbName

##############################################################################################################################


import inspect
from types import FunctionType, MethodType
from typing import Iterable

#############################################################################################################

def toIterable(
    *items,
    ignoreString: bool = True
):
    """
    Function to make item iterable
    """
    iterableItems = []
    for item in items:
        if hasattr(item, '__iter__'):
            iterableItem = [item] if isinstance(item, (str, bytes)) and ignoreString else item
        else:
            iterableItem = [item]
        #yield from iterableItem
        iterableItems.extend(iterableItem)
    return tuple(iterableItems)# if len(iterableItems) > 1 else iterableItems[0]

#############################################################################################################

def itemReplacer(
    dict: dict,
    items: object
):
    """
    Function to replace item using dictionary lookup
    """
    itemList = toIterable(items, ignoreString = False)

    itemList_new = [dict.get(item, item) for item in itemList]

    if isinstance(items, list):
        return itemList_new
    if isinstance(items, tuple):
        return tuple(itemList_new)
    if isinstance(items, (int, float, bool)):
        return itemList_new[0]
    if isinstance(items, str):
        return str().join(itemList_new)


def findKey(
    dict: dict,
    targetValue: object
):
    """
    Find key from dictionary
    """
    for key, value in dict.items():
        if value == targetValue:
            return key

#############################################################################################################

def getNamesFromMethod(
    method: object
):
    """
    Function to get className and methodName from classmethod
    """
    if type(method) not in [FunctionType, MethodType]:
        raise Exception("Only accept classmethod or function")
    qualName = method.__qualname__
    className, methodName = qualName.split('.') if '.' in qualName else (None, qualName)
    return className, methodName


def getClassFromMethod(
    method: object
):
    """
    Function to get class from classmethod
    """
    className = getNamesFromMethod(method)[0]
    return inspect.getmodule(method).__dict__[className]

#############################################################################################################

def runEvents(
    events: Iterable
):
    """
    Function to run events
    """
    if isinstance(events, dict):
        for event, param in events.items():
            event(*toIterable(param if param is not None else ())) if event is not None else None
    else:
        for event in iter(events):
            event() if event is not None else None

#############################################################################################################


import os
import sys
import platform
import re
from packaging import version
from enum import Enum

from .text import rawString
from .path import normPath
from .cmd import runCMD

#############################################################################################################

def isVersionSatisfied(
    currentVersion: str,
    versionReqs: str
):
    """
    Check if the version requirements are satisfied
    """
    if versionReqs is None:
        return True
    versionReqs = versionReqs.split(',') if isinstance(versionReqs, str) else list(versionReqs)
    results = []
    for versionReq in versionReqs:
        splitVersionReq = re.split('!|=|>|<', versionReq)
        requiredVersion = splitVersionReq[-1].strip()
        req = versionReq[:len(versionReq) - len(splitVersionReq[-1])].strip()
        if req == "":
            results.append(version.parse(currentVersion) == version.parse(requiredVersion))
        elif req == "==":
            results.append(version.parse(currentVersion) == version.parse(requiredVersion))
        elif req == "!=":
            results.append(version.parse(currentVersion) != version.parse(requiredVersion))
        elif req == ">=":
            results.append(version.parse(currentVersion) >= version.parse(requiredVersion))
        elif req == "<=":
            results.append(version.parse(currentVersion) <= version.parse(requiredVersion))
        elif req == ">":
            results.append(version.parse(currentVersion) > version.parse(requiredVersion))
        elif req == "<":
            results.append(version.parse(currentVersion) < version.parse(requiredVersion))
    return True if False not in results else False


def isSystemSatisfied(
    systemReqs: str
):
    """
    Check if the system requirements are satisfied
    """
    if systemReqs is None:
        return True
    systemReqs = systemReqs.split(';') if isinstance(systemReqs, str) else list(systemReqs)
    results = []
    for systemReq in systemReqs:
        splitSystemReq = re.split('!|=|>|<', systemReq)
        requiredSystem = splitSystemReq[-1].strip()
        req = systemReq[len(splitSystemReq[0]) : len(systemReq) - len(splitSystemReq[-1])].strip()
        if req == "==":
            results.append(sys.platform == eval(requiredSystem))
        if req == "!=":
            results.append(sys.platform != eval(requiredSystem))
        return True if False not in results else False

#############################################################################################################

class envType(Enum):
    System = "System"
    User = "User"
    Temp = "Temp"


def setEnvVar(
    variable: str,
    value: str,
    type: envType = envType.Temp,
    affectOS: bool = True
):
    """
    Set environment variable
    """
    value = rawString(value)

    if type == envType.System:
        if platform.system() == 'Windows':
            command = ";".join([
                f"`$current = [Environment]::GetEnvironmentVariable('{variable}', 'Machine')",
                f"`$new = '{value};' + `$current",
                f"[Environment]::SetEnvironmentVariable('{variable}', `$new, 'Machine')",
            ])
            runCMD(
                args = [
                    f'powershell.exe',
                    f'Start-Process powershell -ArgumentList "-Command {command}" -Verb RunAs -Wait',
                ],
                shell = True,
                env = os.environ,
            )
        if platform.system() == 'Linux':
            with open('/etc/environment', 'a') as f:
                f.write(f'\n{variable}="{value}"\n')

    if type == envType.User:
        if platform.system() == 'Windows':
            runCMD(
                args = [
                    f'for /f "usebackq tokens=2,*" %A in (`reg query "HKEY_CURRENT_USER\\Environment" /v "{variable}"`) do set userVAR=%B',
                    f'if defined userVAR (setx "{variable}" "{value}{os.pathsep}%userVAR%") else (setx "{variable}" "{value}")'
                ],
                shell = True,
                env = os.environ,
            )
        if platform.system() == 'Linux':
            shell = os.environ.get('SHELL', '/bin/bash')
            if 'bash' in shell:
                config_file = os.path.expanduser('~/.bashrc')
            elif 'zsh' in shell:
                config_file = os.path.expanduser('~/.zshrc')
            else:
                config_file = os.path.expanduser('~/.profile')
            with open(config_file, 'a') as f:
                f.write(f'\nexport {variable}="{value}"\n')

    if type == envType.Temp or affectOS:
        EnvValue = os.environ.get(variable)
        if EnvValue is not None and normPath(value, 'Posix') not in [normPath(value, 'Posix') for value in EnvValue.split(os.pathsep)]:
            EnvValue = f'{value}{os.pathsep}{EnvValue}' #EnvValue = f'{EnvValue}{os.pathsep}{value}'
        else:
            EnvValue = value
        os.environ[variable] = EnvValue

#############################################################################################################


import os
import platform
import socket
import requests
import urllib
import hashlib
import json
from packaging import version
from github import Github
from pathlib import Path
from enum import Enum
from typing import Union, Optional, Tuple, Any

from .utils import toIterable
from .path import normPath
from .text import getSystemEncoding
from .cmd import runCMD

#############################################################################################################

def isPortAvailable(port: int, host: str = '127.0.0.1', protocol: str = 'tcp'):
    """
    Check whether port is available
    """
    if protocol == 'tcp':
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    elif protocol == 'udp':
        sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    else:
        raise ValueError("Protocol must be 'tcp' or 'udp'")
    sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
    try:
        sock.bind((host, port))
        sock.close()
        return True
    except (socket.error, OSError):
        return False


def findAvailablePorts(port_range: tuple, host: str = '127.0.0.1', protocol: str = 'tcp'):
    """
    Find available ports
    """
    available_ports = []
    port_range = (port_range, port_range + 1) if isinstance(port_range, int) else (port_range[0], port_range[1] + 1)
    for port in range(*port_range):
        if isPortAvailable(port, host, protocol):
                available_ports.append(port)
    return available_ports


def freePort(port: int):
    """
    Free port
    """
    if isPortAvailable(port):
        return
    netStat = runCMD('netstat -aon|findstr "%s"' % port)
    for line in str(netStat).splitlines():
        line = line.strip()
        runCMD(f'taskkill /T /F /PID {line.split(" ")[-1]}') if line.startswith("TCP") else None

#############################################################################################################

class requestManager(Enum):
    """
    Manage request
    """
    Post = 0
    Get = 1
    #Head = 2

    def _request(self,
        reqMethod: str,
        protocol: str,
        host: str,
        port: int,
        pathParams: Union[str, list[str], None],
        queryParams: Union[str, list[str], None],
        headers: Optional[dict],
        data: Any,
        stream: bool,
        **kwargs
    ):
        pathParams = "/".join([str(pathParam) for pathParam in toIterable(pathParams)] if pathParams else [])
        queryParams = "&".join([str(queryParam) for queryParam in toIterable(queryParams)] if queryParams else [])
        response = requests.request(
            method = reqMethod,
            url = f"{protocol}://{host}:{port}"
            + (f"/{pathParams}" if len(pathParams) > 0 else "")
            + (f"?{queryParams}" if len(queryParams) > 0 else ""),
            headers = headers,
            data = data if isinstance(data, str) else (json.dumps(data) if data is not None else None),
            stream = stream,
            **kwargs
        )
        #assert response.status_code == 200
        return response

    def request(self,
        protocol: str = "http",
        host: str = "127.0.0.1",
        port: int = 8080,
        pathParams: Union[str, list[str], None] = None,
        queryParams: Union[str, list[str], None] = None,
        headers: Optional[dict] = None,
        data: Any = None,
        stream: bool = False,
        **kwargs
    ): 
        if self == self.Post:
            reqMethod = 'POST'
        if self == self.Get:
            reqMethod = 'GET'
        return self._request(reqMethod, protocol, host, port, pathParams, queryParams, headers, data, stream, **kwargs)

    def response(self,
        protocol: str,
        host: str,
        port: int,
        pathParams: Union[str, list[str], None] = None,
        queryParams: Union[str, list[str], None] = None,
        headers: Optional[dict] = None,
        data: Any = None,
        stream: bool = False,
        decodeUnicode: bool = True,
        **kwargs,
    ):
        if self == self.Post:
            reqMethod = 'POST'
        if self == self.Get:
            reqMethod = 'GET'
        with self._request(reqMethod, protocol, host, port, pathParams, queryParams, headers, data, stream, **kwargs) as response:
            if response.status_code == 200:
                for parsed_content, status_code in responseParser(response, stream, decodeUnicode):
                    yield parsed_content, status_code
            else:
                yield "Request failed", response.status_code
                return


def simpleRequest(
    reqMethod: requestManager,
    protocol: str,
    host: str,
    port: int,
    pathParams: Union[str, list[str], None] = None,
    queryParams: Union[str, list[str], None] = None,
    headers: Optional[dict] = None,
    data: Any = None,
    stream: bool = False,
    *keys
):
    with reqMethod.request(protocol, host, port, pathParams, queryParams, headers, data, stream) as response:
        encodedResponse = response.json()
        result = tuple([encodedResponse.get(key, {}) for key in keys]) if keys else encodedResponse
        return result

#############################################################################################################

def _download(
    downloadURL: str,
    downloadPath: str,
):
    with urllib.request.urlopen(downloadURL) as source, open(downloadPath, mode = "wb") as output:
        totalLength = int(source.info().get("content-Length"))
        while True:
            buffer = source.read(8192)
            if not buffer:
                break
            output.write(buffer)
            yield len(buffer), totalLength


def _download_aria(
    downloadURL: str,
    downloadPath: str,
    createNewConsole: bool = False
):
    runCMD(
        args = [
            'aria2c',
            f'''
            {('cmd.exe /c start ' if platform.system() == 'Windows' else 'x-terminal-emulator -e ') if createNewConsole else ''}
            aria2c "{downloadURL}" --dir="{Path(downloadPath).parent.as_posix()}" --out="{Path(downloadPath).name}" -x6 -s6 --file-allocation=none --force-save=false
            '''
        ]
    )


def downloadFromURL(
    downloadURL: str,
    downloadDir: str,
    fileName: str,
    fileFormat: str,
    sha: Optional[str],
    createNewConsole: bool = False
) -> Tuple[Union[bytes, str], str]:
    """
    Downloads a file from a given URL and saves it to a specified directory
    """
    fileBytes = None
    isDownloadNeeded = True

    downloadName = fileName + (fileFormat if '.' in fileFormat else f'.{fileFormat}')
    downloadPath = normPath(Path(downloadDir).joinpath(downloadName).absolute())

    if Path(downloadPath).exists():
        if Path(downloadPath).is_file() and sha is not None:
            with open(downloadPath, mode = "rb") as f:
                fileBytes = f.read()
            if len(sha) == 40:
                SHA_Current = hashlib.sha1(fileBytes).hexdigest()
            if len(sha) == 64:
                SHA_Current = hashlib.sha256(fileBytes).hexdigest()
            isDownloadNeeded = True if SHA_Current != sha else False
        else:
            os.remove(downloadPath)
            os.makedirs(downloadDir, exist_ok = True)

    if isDownloadNeeded:
        try:
            _download_aria(downloadURL, downloadPath, createNewConsole)
        except:
            iter(_download(downloadURL, downloadPath))
        finally:
            fileBytes = open(downloadPath, mode = "rb").read() if Path(downloadPath).exists() else None

    if fileBytes is None:
        raise Exception('Download Failed!')

    return fileBytes, downloadPath

#############################################################################################################

def checkUpdateFromGithub(
    repoOwner: str = ...,
    repoName: str = ...,
    fileName: str = ...,
    fileFormat: str = ...,
    currentVersion: str = ...,
    accessToken: Optional[str] = None,
):
    """
    Check if there is an update available on Github
    """
    try:
        PersonalGit = Github(accessToken)
        Repo = PersonalGit.get_repo(f"{repoOwner}/{repoName}")
        latestVersion = Repo.get_tags()[0].name
        latestRelease = Repo.get_latest_release() #latestRelease = Repo.get_release(latestVersion)
        for Index, Asset in enumerate(latestRelease.assets):
            if Asset.name == f"{fileName}.{fileFormat}":
                IsUpdateNeeded = True if version.parse(currentVersion) < version.parse(latestVersion) else False
                downloadURL = Asset.browser_download_url #downloadURL = f"https://github.com/{repoOwner}/{repoName}/releases/download/{latestVersion}/{fileName}.{fileFormat}"
                VersionInfo = latestRelease.body
                return IsUpdateNeeded, downloadURL, VersionInfo
            elif Index + 1 == len(latestRelease.assets):
                raise Exception(f"No file found with name {fileName}.{fileFormat} in the latest release")

    except Exception as e:
        print(f"Error occurred while checking for updates: \n{e}")

#############################################################################################################


import sys
from loguru import logger
from pathlib import Path
from enum import Enum
from typing import Optional

#############################################################################################################

class loggerLevel(Enum):
    DEBUG = "DEBUG"
    INFO = "INFO"
    WARNING = "WARNING"
    ERROR = "ERROR"
    CRITICAL = "CRITICAL"

    def no(self):
        return logger.level(self.name).no


_format = " | ".join([
    '<green>{time:YYYY-MM-DD HH:mm:ss.SSS}</green>',
    '<level>{level: <8}</level>',
    #'<magenta>{process}</magenta>:<yellow>{thread}</yellow>',
    '<cyan>{name}</cyan>:<cyan>{function}</cyan>:<cyan>{line}</cyan> - <level>{message}</level>',
])


class loggerManager:
    """
    Manage logger
    """
    def __init__(self):
        self.isLoggerInitialized = False

    def createLogger(self,
        name: str,
        level: loggerLevel = loggerLevel.INFO,
        format: str = _format,
        outputPath: Optional[str] = None,
        rotation: str = "10 MB",
        useStdIO: bool = True,
    ):
        if not self.isLoggerInitialized:
            logger.remove()
            self.isLoggerInitialized = True

        filter = lambda record: record["extra"].get("name") == name

        stream = (sys.stdout or sys.stderr or None) if useStdIO else None
        if stream:
            logger.add(
                stream,
                format = format,
                level = level.value,
                filter = filter,
            )

        if outputPath:
            dir = Path(outputPath).parent
            dir.mkdir(parents = True, exist_ok=True)

            logger.add(
                Path(outputPath).as_posix(),
                level = level.value,
                format = format,
                backtrace = True,
                diagnose = True,
                enqueue = True,
                rotation = rotation,
                filter = filter,
            )

        return logger.bind(name = name)

#############################################################################################################


# -*- coding:utf-8 -*-

from setuptools import setup, find_packages

##############################################################################################################################

with open('./README.md', encoding = 'utf-8') as f:
    LongDescription = f.read()

##############################################################################################################################

setup(
    name = "PyEasyUtils",
    version = '0.3.6',
    description = 'A simple python util library',
    long_description = LongDescription,
    long_description_content_type = 'text/markdown',
    license = 'MIT',
    author = "Spr_Aachen",
    author_email = '2835946988@qq.com',
    url = 'https://github.com/Spr-Aachen/PyEasyUtils',
    project_urls = {
        'Source Code': 'https://github.com/Spr-Aachen/PyEasyUtils',
        'Bug Tracker': 'https://github.com/Spr-Aachen/PyEasyUtils/issues',
    },
    packages = find_packages(
        where = '.',
        exclude = ()
    ),
    include_package_data = True,
    install_requires = [
        "psutil",
        "loguru",
        "polars",
        "sqlalchemy",
        "PyGithub",
    ],
    classifiers = [
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
    ]
)

##############################################################################################################################


import os
import sys
import platform
import shlex
import subprocess
import asyncio
from pathlib import Path
from typing import Union, Optional, List

from .utils import toIterable
from .path import normPath, getFileInfo
from .text import getSystemEncoding, removeLF
from .log import loggerManager

#############################################################################################################

class asyncSubprocessManager:
    """
    Manage subprocess of commands (async version)
    """
    def __init__(self,
        shell: bool = False,
        encoding: Optional[str] = None,
    ):
        self.shell = shell
        self.encoding = encoding or getSystemEncoding()

        self.subprocesses: List[asyncio.subprocess.Process] = []

        self.isWindowsSystem = platform.system() == 'Windows'
        #asyncio.set_event_loop(asyncio.ProactorEventLoop()) if self.isWindowsSystem and isinstance(asyncio.get_event_loop(), asyncio.SelectorEventLoop) else None

    async def _create(self, arg: Union[List[str], str], merge: bool, env: Optional[os._Environ] = None):
        limit = 1024 * 1024
        if self.shell == False:
            arg = shlex.split(arg) if isinstance(arg, str) else arg
            process = await asyncio.create_subprocess_exec(
                *arg,
                stdout = asyncio.subprocess.PIPE,
                stderr = asyncio.subprocess.STDOUT,
                env = env or os.environ,
                creationflags = subprocess.CREATE_NO_WINDOW if self.isWindowsSystem else 0,
                text = False,
                limit = limit,
            )
        else:
            arg = shlex.join(arg) if isinstance(arg, list) else arg
            argBuffer = (f'{arg}\n' if not arg.endswith('\n') else arg).encode(self.encoding)
            if platform.system() == 'Windows':
                shellArgs = ['cmd']
            if platform.system() == 'Linux':
                shellArgs = ['bash', '-s']
            process = await asyncio.create_subprocess_exec(
                *shellArgs,
                stdin = asyncio.subprocess.PIPE,
                stdout = asyncio.subprocess.PIPE,
                stderr = asyncio.subprocess.STDOUT,
                env = env or os.environ,
                creationflags = subprocess.CREATE_NO_WINDOW if self.isWindowsSystem else 0,
                text = False,
                limit = limit,
            ) if self.subprocesses.__len__() == 0 or not merge else self.subprocesses[-1]
            process.stdin.write(argBuffer)
            await process.stdin.drain()
        self.subprocesses.append(process)

    async def create(self, args: Union[list[Union[list, str]], str], merge: bool = True, env: Optional[os._Environ] = None):
        for arg in toIterable(args):
            await self._create(arg, merge, env)
            process = self.subprocesses[-1]
            if self.shell == False:
                pass
            else:
                process.stdin.close() if (merge and self.subprocesses.__len__() == toIterable(args).__len__()) or not merge else None

    async def _getOutputLines(self, process: asyncio.subprocess.Process, showProgress: bool = True, logPath: Optional[str] = None):
        logger = loggerManager().createLogger(
            name = "cmd",
            format = "{message}",
            outputPath = logPath,
            useStdIO = False
        ) if logPath is not None else None

        while True:
            line = await process.stdout.readline()
            if not line:
                break
            yield line
            lineString = line.decode(self.encoding, errors = 'replace')
            sys.stdout.write(lineString) if showProgress and sys.stdout is not None else None
            logger.info(removeLF(lineString, removeAll = False)) if logger is not None else None
            if process.returncode is not None:
                break

    async def monitor(self, showProgress: bool = True, logPath: Optional[str] = None):
        for process in self.subprocesses:
            async for line in self._getOutputLines(process, showProgress, logPath):
                yield line, b''
            await process.wait()
            if process.returncode != 0:
                yield b'', b"error occurred, please check the logs for full command output."

    async def close(self):
        for process in self.subprocesses:
            try:
                if process.returncode is None:
                    process.terminate()
                    await asyncio.sleep(0.1)
                    if process.returncode is None:
                        process.kill()
                await process.wait()
                if hasattr(process, '_transport') and process._transport:
                    process._transport.close()
            except:
                pass
        self.subprocesses.clear()


class subprocessManager:
    """
    Manage subprocess of commands (synchronous wrapper)
    """
    def __init__(self, shell: bool = False, encoding: Optional[str] = None):
        self._asyncManager = asyncSubprocessManager(shell, encoding)

        self._hasLoop = False
        self._loop: Optional[asyncio.AbstractEventLoop] = None
        self._isWindowsSystem = self._asyncManager.isWindowsSystem

    def _create_event_loop(self):
        if self._isWindowsSystem:
            try:
                loop = asyncio.ProactorEventLoop()
            except AttributeError:
                loop = asyncio.new_event_loop()
            asyncio.set_event_loop(loop)
        else:
            loop = asyncio.new_event_loop()
            asyncio.set_event_loop(loop)
        self._hasLoop = True
        self._loop = loop
        return loop

    @property
    def subprocesses(self):
        return self._asyncManager.subprocesses

    @property
    def encoding(self):
        return self._asyncManager.encoding

    def _run_async(self, coro):
        try:
            loop = asyncio.get_event_loop()
            if loop.is_closed():
                raise RuntimeError
        except RuntimeError:
            loop = self._create_event_loop()
        else:
            if self._isWindowsSystem and not isinstance(loop, asyncio.ProactorEventLoop):
                loop = self._create_event_loop()
        return loop.run_until_complete(coro)

    def create(self, args: Union[List[Union[List, str]], str], merge: bool = True, env: Optional[os._Environ] = None):
        return self._run_async(self._asyncManager.create(args, merge, env))

    def monitor(self, showProgress: bool = True, logPath: Optional[str] = None, realTime: bool = True):
        if realTime:
            async_monitor = self._asyncManager.monitor(showProgress, logPath)
            try:
                while True:
                    try:
                        yield self._run_async(async_monitor.__anext__())
                    except StopAsyncIteration:
                        break
            finally:
                try:
                    self._run_async(async_monitor.aclose())
                except AttributeError:
                    pass
        else:
            async def collect():
                results = []
                async for o, e in self._asyncManager.monitor(showProgress, logPath):
                    results.append((o, e))
                return results
            for o, e in self._run_async(collect()):
                yield o, e

    def close(self):
        if self._asyncManager.subprocesses:
            self._run_async(self._asyncManager.close())
        if self._hasLoop and self._loop is not None and not self._loop.is_closed():
            self._loop.close()
        self._loop = None
        self._hasLoop = False

    def result(self, decodeResult: Optional[bool] = None, showProgress: bool = True, logPath: Optional[str] = None):
        try:
            output, error = (bytes(), bytes())
            for o, e in self.monitor(showProgress, logPath, realTime = False):
                output += o
                error += e
        finally:
            output, error = output.strip(), error.strip()
            output, error = output.decode(self.encoding, errors = 'ignore') if decodeResult else output, error.decode(self.encoding, errors = 'ignore') if decodeResult else error
            returncode = self.subprocesses[-1].returncode
            self.close()
            return None if output in ('', b'') else output, None if error in ('', b'') else error, returncode


def runCMD(
    args: Union[list[Union[list, str]], str],
    merge: bool = True,
    shell: bool = False,
    env: Optional[os._Environ] = None,
    decodeResult: Optional[bool] = None,
    showProgress: bool = True,
    logPath: Optional[str] = None
):
    """
    Run command
    """
    manageSubprocess = subprocessManager(shell)
    manageSubprocess.create(args, merge, env)
    return manageSubprocess.result(decodeResult, showProgress, logPath)

#############################################################################################################

def mkPyFileCommand(filePath: str, **kwargs):
    args = " ".join([f"--{name} {value}" for name, value in kwargs.items()])
    command = ('' if not filePath.strip().endswith('.py') else 'python ') + '"%s" %s' % (filePath, args)
    return command

#############################################################################################################

def runScript(
    *commands: str,
    scriptPath: Optional[str]
):
    """
    Run a script with bash or bat
    """
    if platform.system() == 'Linux':
        scriptPath = Path.cwd().joinpath('Bash.sh') if scriptPath is None else normPath(scriptPath)
        with open(scriptPath, 'w') as bashFile:
            commands = "\n".join(toIterable(commands))
            bashFile.write(commands)
        os.chmod(scriptPath, 0o755) # 给予可执行权限
        subprocess.Popen(['bash', scriptPath])
    if platform.system() == 'Windows':
        scriptPath = Path.cwd().joinpath('Bat.bat') if scriptPath is None else normPath(scriptPath)
        with open(scriptPath, 'w') as BatFile:
            commands = "\n".join(toIterable(commands))
            BatFile.write(commands)
        subprocess.Popen([scriptPath], creationflags = subprocess.CREATE_NEW_CONSOLE)


def bootWithScript(
    programPath: str = ...,
    delayTime: int = 3,
    scriptPath: Optional[str] = None
):
    """
    Boot the program with a script
    """
    if platform.system() == 'Linux':
        _, isFileCompiled = getFileInfo(programPath)
        runScript(
            '#!/bin/bash',
            f'sleep {delayTime}',
            f'./"{programPath}"' if isFileCompiled else f'python3 "{programPath}"',
            'rm -- "$0"',
            scriptPath = scriptPath
        )
    if platform.system() == 'Windows':
        _, isFileCompiled = getFileInfo(programPath)
        runScript(
            '@echo off',
            f'ping 127.0.0.1 -n {delayTime + 1} > nul',
            f'start "Programm Running" "{programPath}"' if isFileCompiled else f'python "{programPath}"',
            'del "%~f0"',
            scriptPath = scriptPath
        )

#############################################################################################################


import os
import configparser
from datetime import datetime
from pathlib import Path
from typing import Optional

from .path import normPath

#############################################################################################################

class configManager:
    """
    Manage config through ConfigParser
    """
    def __init__(self,
        configPath: Optional[str] = None
    ):
        self.configParser = configparser.ConfigParser()
        self.configPath = normPath(Path(os.getcwd()).joinpath('config_%s.ini' % datetime.now().strftime("%Y-%m-%d_%H-%M-%S"))) if configPath == None else configPath
        if Path(self.configPath).exists():
            self.configParser.read(self.configPath, encoding = 'utf-8')
        else:
            configDir = Path(self.configPath).parent.as_posix()
            os.makedirs(configDir, exist_ok = True) if Path(configDir).exists() == False else None
            with open(self.configPath, 'w'):
                pass
            self.configParser.clear()

    def parser(self):
        return self.configParser

    def editConfig(self,
        section: str = ...,
        option: str = ...,
        value: str = ...,
        configParser: Optional[configparser.ConfigParser] = None
    ):
        configParser = self.parser() if configParser == None else configParser
        try:
            configParser.add_section(section)
        except:
            pass
        configParser.set(section, option, value)
        with open(self.configPath, 'w', encoding = 'utf-8') as Config:
            configParser.write(Config)

    def remove(self,
        section: str = ...,
        option: Optional[str] = None,
        configParser: Optional[configparser.ConfigParser] = None
    ):
        configParser = self.parser() if configParser == None else configParser
        try:
            configParser.remove_option(section, option) if option is not None else configParser.remove_section(section)
        except:
            pass
        with open(self.configPath, 'w', encoding = 'utf-8') as Config:
            configParser.write(Config)

    def getValue(self,
        section: str = ...,
        option: str = ...,
        initValue: Optional[str] = None,
        configParser: Optional[configparser.ConfigParser] = None
    ):
        configParser = self.parser() if configParser == None else configParser
        try:
            value = configParser.get(section, option)
        except:
            if initValue != None:
                self.editConfig(section, option, initValue, configParser)
                return initValue
            else:
                return None #raise Exception("Need initial value")
        else:
            return value

#############################################################################################################